//  Ac3d - 3D objects
//
//	Vamos Automotive Simulator
//  Copyright (C) 2003 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _AC3D_H_
#define _AC3D_H_

#include <vamos/geometry/Three_Vector.h>
#include <vamos/geometry/Three_Matrix.h>

#include <GL/gl.h>

#include <fstream>
#include <string>
#include <vector>

namespace Vamos_Geometry
{
  class Gl_Texture_Image;

  class Ac3d_Exception
  {
	std::string m_message;
  public:
	Ac3d_Exception (std::string message) : m_message (message) {};
	std::string message () const { return m_message; }
  };

  class No_File : public Ac3d_Exception 
  {
  public:
	No_File (std::string message) : Ac3d_Exception (message) {};
  };

  class Not_An_Ac3d_File : public Ac3d_Exception
  {
  public:
	Not_An_Ac3d_File (std::string message) : Ac3d_Exception (message) {};
  };

  class Malformed_Ac3d_File : public Ac3d_Exception
  {
  public:
	Malformed_Ac3d_File (std::string message) : Ac3d_Exception (message) {};
  };

  //===========================================================================
  class Ac3d_Material
  {
	friend class Ac3d_Test;

	std::string m_name;
	double m_color [3];
	double m_ambient [3];
	double m_emission [3];
	double m_specular [3];
	double m_shininess;
	double m_transparency;

  public:
	Ac3d_Material (std::string name_in,
				   double color_in [3],
				   double ambient_in [3],
				   double emission_in [3],
				   double specular_in [3],
				   double shininess_in,
				   double transparency_in);

	double color (size_t index) const;
	double ambient (size_t index) const;
	double emission (size_t index) const;
	double specular (size_t index) const;
	double shininess () const { return m_shininess; }
	double transparency () const { return m_transparency; }
  };

  //===========================================================================
  class Ac3d_Surface
  {
  public:
	struct Vertex
	{
	  Vertex (const Three_Vector* coords, 
			  const Three_Vector* normal,
			  double tex_x, double tex_y);
	  const Three_Vector* coords;
	  const Three_Vector* normal;
	  double tex_x;
	  double tex_y;
	};

	enum Type { POLYGON, LINE, CLOSED_LINE };

  private:
	const Ac3d_Material* mp_material;
	std::vector <const Vertex*> m_vertices;
	Three_Vector m_normal;

	Type m_type;
	bool m_shaded;
	bool m_two_sided;

	double m_scale;
	const Three_Vector* mp_offset;
	const Three_Matrix* mp_rotation;

  public:
	Ac3d_Surface (std::string type, 
				  double scale,
				  const Three_Vector* offset,
				  const Three_Matrix* rotation);
	~Ac3d_Surface ();

	void set_material (const Ac3d_Material* mat) { mp_material = mat; }
	const Ac3d_Material* get_material () const { return mp_material; }

	void add_vertices (const std::vector <const Vertex*>& verts);
	const std::vector <const Vertex*> get_vertices () const
	{ 
	  return m_vertices; 
	}

	void add_normal (const Three_Vector& norm) { m_normal = norm; }

	Type get_type () const { return m_type; }
	bool is_shaded () const { return m_shaded; }
	bool is_two_sided () const { return m_two_sided; }

	void build () const;
  };

  //===========================================================================
  class Ac3d_Object
  {
	friend class Ac3d_Test;

  public:
	struct Vertex
	{
	  Vertex (const Three_Vector* coords_in) : coords (coords_in) {};
	  ~Vertex () { delete coords; }
	  const Three_Vector* coords;
	  Three_Vector normal;
	};

  private:
	std::string m_type;
	std::string m_name;
	std::string m_data;
	std::string m_url;

	double m_scale;
	Three_Matrix m_rotation;
	Three_Vector m_location;

	Gl_Texture_Image* mp_texture;
	double m_texture_repeat [2];

	std::vector <const Ac3d_Object*> m_kids;
	std::vector <Vertex*> m_vertices;
	std::vector <const Ac3d_Surface*> m_surfaces;

  public:
	Ac3d_Object (std::string type, double scale, 
				 const Three_Vector& translation,
				 const Three_Vector& rotation);
	~Ac3d_Object ();

	void set_name (std::string name) { m_name = name; }
	void set_data (std::string data) { m_data = data; }
	void set_texture_image (std::string file);
	void set_texture_repeat (double x, double y);

	double get_scale () const { return m_scale; }
	void set_rotation (const Three_Matrix& rot);
	const Three_Matrix* get_rotation () const { return &m_rotation; }
	void set_location (const Three_Vector& loc);
	const Three_Vector* get_location () const { return &m_location; }

	void set_url (std::string url) { m_url = url; }
	void add_kid (const Ac3d_Object* obj) { m_kids.push_back (obj); }
	void add_vertex (double x, double y, double z);
	const Three_Vector* get_vertex (size_t index) const;
	void add_normal (size_t index, const Three_Vector& norm);
	const Three_Vector* get_normal (size_t index) const;
	void add_surface (const Ac3d_Surface* surf);

	void build () const;
  };

  //===========================================================================
  class Ac3d
  {
	friend class Ac3d_Test;

	std::string m_file;
	int m_version;

	std::vector <const Ac3d_Material*> m_materials;
	std::vector <const Ac3d_Object*> m_objects;

	double m_scale;
	Three_Vector m_translation;
	Three_Vector m_rotation; // anlgle-axis representation

	void read_header (std::ifstream& is);
	const Ac3d_Material* read_material (std::ifstream& is);
	const Ac3d_Object* read_object (std::ifstream& is, double scale,
									const Three_Vector& translation,
									const Three_Vector& rotation);
	const Ac3d_Surface* read_surface (std::ifstream& is, Ac3d_Object* obj);

  public:
	Ac3d (std::string file, double scale, 
		  const Three_Vector& rotation,
		  const Three_Vector& translation);
	~Ac3d ();

	GLuint build ();
  };
}

#endif // not _AC3D_H_
